/* Generic i386.
   Written by Pieter J. Schoenmakers <tiggr@ics.ele.tue.nl>

   Copyright (C) 1996-1998 Pieter J. Schoenmakers.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: i386.h,v 1.8 1998/03/30 11:13:58 tiggr Exp $  */

#define BUILTIN_RETURN_TYPE  \
union						\
{						\
  struct { tom_int i; } i;			\
  struct { tom_long l; } l;			\
  struct { void *p; } p;			\
  struct { tom_int a, b; char d[10]; } d;	\
  struct { tom_int a, b; char f[10]; } f;	\
}

/* The return type of a function that returns using __builtin_return.  */
#define GENERIC_RETURN_TYPE  double

#define RETURN_RETRIEVE_DOUBLE(RESULT)  \
  ({								\
    double d;							\
								\
    __asm__ volatile ("fldt (%0)" : : "g" ((RESULT)->d.d));	\
    __asm__ volatile ("fstpl (%0)" : : "g" (&d));		\
    d;								\
  })

#define RETURN_RETRIEVE_FLOAT(RESULT)  RETURN_RETRIEVE_DOUBLE (RESULT)

#define RETURN_SET_DOUBLE(RESULT, VALUE)  \
  do								\
    {								\
      double d = (VALUE);					\
      __asm__ volatile ("fldl (%0)" : : "g" (&d));		\
      __asm__ volatile ("fstpt (%0)" : : "g" ((RESULT)->d.d));	\
    } while (0)

#define RETURN_SET_FLOAT(RESULT, VALUE)  RETURN_SET_DOUBLE (RESULT, VALUE)

#define TGC_MARK_REGISTERS()  \
  do							\
    {							\
      /* XXX No need to mark the 0.001 registers...  */	\
    } while (0)

/* The macros for building the arguments to __builtin_apply (i.e. most
   of the rest of this file) depend on the fact that _all_ arguments
   are passed on the stack.  In fact, this code equals the m68k-next
   code where the same applies.  */

/* Macro local to this target file to emit a value of some type to the
   ARGS being built.  */
#define APPLY_ARGS_EMIT_VALUE(TYPE, VALUE)  \
({								\
    TYPE *addr = (void *) ((char *) args->stack + offset);	\
    *addr = (VALUE);						\
    offset += sizeof (TYPE);					\
    addr;})

/* The size of the registers in the ARGS to __builtin_apply.
   XXX This is 0, as all arguments are passed on the stack.  */
#define APPLY_ARGS_REG_SIZE  0

/* Start building the ARGS to __builtin_apply.
   This should be called init_cumulative_args :-) */
#define APPLY_ARGS_START(SEL)  \
  do			\
    {			\
      int offset = 0

/* Return an estimate, which is guaranteed to be large enough, for the
   size of the stack arguments to __builtin_apply_args.

   Worst case is the implicit first arguments SELF and CMD, followed
   by all double arguments, followed by the tuple return value
   pointers.  */
#define APPLY_ARGS_STACK_SIZE(SEL)  \
  (2 * sizeof (void *) + sizeof (tom_double) * (SEL)->in->num	\
   + sizeof (void *) * (SEL)->out->num)

/* Add a byte value to the ARGS being built.  */
#define APPLY_ARGS_EMIT_BYTE(VALUE)  \
  APPLY_ARGS_EMIT_INT ((tom_int) (VALUE))

/* Add a char value to the ARGS being built.  */
#define APPLY_ARGS_EMIT_CHAR(VALUE)  \
  APPLY_ARGS_EMIT_INT ((tom_int) (VALUE))

/* Add a int value to the ARGS being built.  */
#define APPLY_ARGS_EMIT_INT(VALUE)  \
  APPLY_ARGS_EMIT_VALUE (tom_int, VALUE)

/* Add a long value to the ARGS being built.  */
#define APPLY_ARGS_EMIT_LONG(VALUE)  \
  APPLY_ARGS_EMIT_VALUE (tom_long, VALUE)

/* Add a float value to the ARGS being built.  */
#define APPLY_ARGS_EMIT_FLOAT(VALUE)  \
  APPLY_ARGS_EMIT_VALUE (tom_float, VALUE)

/* Add a double value to the ARGS being built.  */
#define APPLY_ARGS_EMIT_DOUBLE(VALUE)  \
  APPLY_ARGS_EMIT_VALUE (tom_double, VALUE)

/* Add a selector to the ARGS being built.  */
#define APPLY_ARGS_EMIT_SELECTOR(VALUE)  \
  APPLY_ARGS_EMIT_POINTER (VALUE)

/* Add a reference to the ARGS being built.  */
#define APPLY_ARGS_EMIT_REFERENCE(VALUE)  \
  APPLY_ARGS_EMIT_POINTER (VALUE)

/* Add a pointer to the ARGS being built.  */
#define APPLY_ARGS_EMIT_POINTER(VALUE)  \
  APPLY_ARGS_EMIT_VALUE (void *, VALUE)

/* Perform any wrapping up of the ARGS needed to perform the actual
   application.  */
#define APPLY_ARGS_COMPLETE()  \
      do ; while (0)

/* Return the actual size of the stack arguments to __builtin_apply_args.  */
#define APPLY_ARGS_ACTUAL_SIZE()  \
      offset

/* Undo anything started by APPLY_ARGS_START.  */
#define APPLY_ARGS_END()  \
    } while (0)
